;; Copyright (C) 2008-2013 Tommi Höynälänmaa
;; Distributed under GNU General Public License version 3,
;; see file doc/GPL-3.


;; *** Linker main program ***


(import (rnrs exceptions)
	(srfi srfi-1)
	(ice-9 pretty-print)
	(th-scheme-utilities stdutils)
	(th-scheme-utilities hrecord))


(define gl-l-all-int-lang '(tree-il-2.2 tree-il-2.0 scheme))

(define gl-l-tree-il-lang '(tree-il-2.2 tree-il-2.0))

(define gl-i-ht-used-size 20000)

(define gl-i-ht-method-defs-size 10000)

(define gl-i-ht-method-decls-size 10000)

(define gl-i-ht-used-decls-size 10000)

(define gl-i-ht-rebound-size 10000)

(define gl-i-ht-decl-types-size 10000)

(define gl-i-ht-globals-by-address-size 10000)

(define gl-i-ht-globals-by-name-size 10000)

(define gl-i-ht-cycles-size 1000)

(define gl-i-ht-goops-classes-size 2000)

(define gl-i-ht-prim-classes-size 500)

(define gl-i-ht-fact-size 500)

(define gl-i-ht-module-indices-size 100)

(define gl-i-ht-lexical-vars-size 10000)

(define gl-i-ht-raw-procs-size 2000)

(define gl-i-ht-equal-size 500)

(define gl-i-ht-equal-objects-size 500)

(define gl-i-ht-equal-contents-size 500)

(define gl-debug '())

(define gl-flag20?)


(define (use-tree-il? linker)
  (memq (hfield-ref linker 's-intermediate-language) gl-l-tree-il-lang))


(define (get-first-free-loc linker module-name)
  (let* ((actual-module-name (get-actual-module-name module-name))
	 (ext (get-pcode-suffix 'body))
	 (module-search-path
	  (hfield-ref linker 'module-search-path))
	 (source-filename (search-file module-search-path
				       actual-module-name ext)))
    (if (string? source-filename)
	(let* ((file-source (theme-open-input-file source-filename))
	       (i-first-free-loc (begin (theme-read file-source)
					(theme-read file-source)
					(theme-read file-source))))
	  (theme-close-input-port file-source)
	  (if (and (integer? i-first-free-loc) (>= i-first-free-loc 0))
	      i-first-free-loc
	      (raise 'invalid-first-free-loc)))
	(raise (list 'module-body-not-found
		     (cons 'module-name module-name))))))


(define (get-interm-file-ext s-int)
  (assert (memq s-int gl-l-all-int-lang))
  (if (memq s-int gl-l-tree-il-lang)
      tree-il-file-ext
      scheme-file-ext))


(define (get-pcode-unit-type source-filename)
  (assert (string? source-filename))
  (let ((ext (get-filename-extension source-filename)))
    (cond
     ((string=? ext "") (raise 'unable-to-compute-unit-type))
     ((string=? ext pcode-proper-program-ext) 'proper-program)
     ((string=? ext pcode-script-ext) 'script)
     (else (raise 'unable-to-compute-unit-type)))))


(define (compile-program-prefix-tree-il linker script?)
  (list
   (if script?
       '(set! (toplevel gl-script?) (const #t))
       '(set! (toplevel gl-script?) (const #f)))       
   (if (hfield-ref linker 'runtime-pretty-backtrace?)
       '(set! (toplevel gl-pretty-backtrace-support?) (const #t))
       '(set! (toplevel gl-pretty-backtrace-support?) (const #f)))))

	       
(define (compile-program-prefix-scheme script?)
  (list
   '(import (guile))
   '(import (srfi srfi-1))
   '(import (rnrs exceptions))
   '(import (rnrs conditions))
   '(import (th-scheme-utilities stdutils))
   (if script?
       '(set! gl-script? #t)
       '(set! gl-script? #f))
   '(set! gl-pretty-backtrace-support? #f)))


(define (not-null-texpr? texpr)
  (and (not (null? texpr)) (not (equal? texpr '(quote ())))))


(define (write-module-body file texpr-list pretty-print?)
  (dwl4 "write-module-body")
  (dvar1-set! texpr-list)
  (if pretty-print?
      (for-each (lambda (texpr)
		  (if (not-null-texpr? texpr)
		      (begin
			(pretty-print texpr file) (newline file))))
		texpr-list)
      (for-each (lambda (texpr)
		  (if (not-null-texpr? texpr)
		      (begin (write texpr file) (newline file))))
		texpr-list))
  (dwl4 "write-module-body EXIT"))


(define (write-target-expr linker file pretty-print? t-expr)
  (if (not-null-texpr? t-expr)
      (let ((lst
	     (if (and (pair? t-expr)
		      (equal? (car t-expr) '_splice))
		 (cdr t-expr)
		 (list t-expr))))
	(guard (exc (else
		     (make-file-exception
		      'error-writing-file
		      (hfield-ref linker 'interm-filename))))
	       (set! gl-debug t-expr)
	       (write-module-body file lst pretty-print?)))))
  

(define (theme-target-compile-program linker lst-entities)
  (dwl2 "theme-target-compile-program")
  (if (use-tree-il? linker)
      (map-in-order (lambda (entity)
		      (let ((result (tc-tree-il-entity linker entity)))
			(if (not-null? result) result '(void))))
		    lst-entities)
      (map-in-order (lambda (entity) (tc-scheme-entity linker entity))
		    lst-entities)))


(define (compile-decl-param-proc-instances linker)
  ;; We compute the list in reversed order in order to optimize.
  (let ((instances (reverse (hfield-ref linker 'decl-proc-instances))))
    (map-in-order (lambda (inst)
		    (assert (= (length inst) 3))
		    (let ((var (car inst))
			  (param-proc (cadr inst))
			  (type-var-values (caddr inst)))
		      (if (use-tree-il? linker)
			  (tc-tree-il-decl-proc-instance linker var
							 param-proc
							 type-var-values)
			  (compile-decl-proc-instance linker var
						      param-proc
						      type-var-values))))
		  instances)))


(define (theme-link-program-imports linker address-env imports)
  (assert (hrecord-is-instance? linker <linker>))
  (assert (hrecord-is-instance? address-env <address-environment>))
  (assert (and (list? imports) (and-map? is-module-name? imports)))
  (for-each
   (lambda (submod)
     (theme-link-unit linker "" submod 'interface address-env '() #f #f))
   imports))


(define (theme-link-body-imports linker address-env mod-name imports)
  (assert (hrecord-is-instance? linker <linker>))
  (assert (hrecord-is-instance? address-env <address-environment>))
  (assert (is-module-name? mod-name))
  (assert (and (list? imports) (and-map? is-module-name? imports)))
  (theme-link-unit linker "" mod-name 'interface address-env '() #f #f)
  (for-each
   (lambda (submod)
     (theme-link-unit linker "" submod 'interface address-env '() #f #f))
   imports))


(define (theme-link-interface-imports linker address-env hierarchy imports)
  (assert (hrecord-is-instance? linker <linker>))
  (assert (hrecord-is-instance? address-env <address-environment>))
  (assert (and (list? hierarchy) (and-map? is-module-name? hierarchy)))
  (assert (and (list? imports) (and-map? is-module-name? imports)))
  (for-each
   (lambda (submod)
     (theme-link-unit linker "" submod 'interface address-env hierarchy #f #f))
   imports))


(define (theme-prelink linker address-env hierarchy l-bodies)
  (assert (hrecord-is-instance? linker <linker>))
  (assert (hrecord-is-instance? address-env <address-environment>))
  (assert (and (list? hierarchy) (and-map? is-module-name? hierarchy)))
  (assert (and (list? l-bodies) (and-map? is-module-name? l-bodies)))
  (for-each
   (lambda (submod)
     (theme-link-unit linker "" submod 'body address-env hierarchy #f #t))
   l-bodies))


;; The output of this procedure updates necessary declared method variables
;; in case the generic procedure where they belong is stripped out.
(define (compile-used-method-decl linker address method-def)
  (assert (is-linker? linker))
  (assert (is-address? address))
  (assert (hrecord-is-instance? method-def <method-definition>))
  (let ((gen-proc (hfield-ref method-def 'gen-proc)))
    (dvar1-set! gen-proc)
    (assert (is-entity? gen-proc))
    ;; We may have to check globals-by-address her.
    (let ((ht-used (hfield-ref linker 'ht-used))
	  (ht-used-decls (hfield-ref linker 'ht-used-decls))
	  (addr-gen-proc (hfield-ref gen-proc 'address)))
      (assert (is-address? addr-gen-proc))
      (if (and (not (address-hash-ref ht-used addr-gen-proc))
	       (address-hash-ref ht-used-decls address))
	(let ((t-var-name (get-target-var-name linker address))
	      (repr-proc (hfield-ref method-def 'procexpr)))
	  (if (use-tree-il? linker)
	      (let ((t-proc (theme-target-tree-il-compile linker repr-proc)))
		`(set! (toplevel ,t-var-name) ,t-proc))
	      (let ((t-proc (theme-target-compile linker repr-proc)))
		(list 'set! t-var-name t-proc))))
	(if (use-tree-il? linker)
	    '(void)
	    '(quote ()))))))
	  

(define (compile-used-method-decls linker)
  (hash-map->list
   (lambda (addr-key method-def)
     (compile-used-method-decl linker addr-key method-def))
   (hfield-ref linker 'ht-used-decls)))


(define (handle-tree-il-splice linker texpr)
  (if (and
       (list? texpr)
       (not-null? texpr)
       (eq? (car texpr) '_splice))
      (tc-tree-il-compile-sequence linker (cdr texpr))
      texpr))


(define (handle-tree-il-splices linker l-texprs)
  (map (lambda (texpr) (handle-tree-il-splice linker texpr)) l-texprs))


(define (filter-void l-texprs)
  (filter (lambda (texpr) (not (equal? texpr '(void)))) l-texprs))


(define (theme-link-unit linker source-filename mod-name unit-type address-env
			 hierarchy program? prelink?)
  (dwl4 "theme-link-unit")
  (assert (hrecord-is-instance? linker <linker>))
  (assert (string? source-filename))
  ;; Module name has to be a list.
  (assert (is-module-name? mod-name))
  (assert (memq unit-type '(proper-program script interface body)))
  (assert (hrecord-is-instance? address-env <address-environment>))
  (assert (and (list? hierarchy) (and-map? is-module-name? hierarchy)))
  (assert (boolean? program?))
  (dw1 "Linking unit: ")
  (dw1 mod-name)
  (dw1 ", ")
  (dwl1 unit-type)
;;  (dwl4 "*1*") (dp)
  (let ((prev-module (hfield-ref linker 'current-module)))
    (hfield-set! linker 'current-module mod-name)
    (cond
     ((and (eq? unit-type 'interface) (member mod-name hierarchy))
      (begin
	;;	(write-error-info mod-name)
	(raise 'cyclic-interface-dependency)))
     ((and prelink? (not (eq? unit-type 'body)))
      (raise 'internal-error-with-prelinking))
     ((and prelink? (member mod-name hierarchy))
      (begin
	;;	(write-error-info mod-name)
	(raise (list 'cyclic-prelink-dependency
		     (cons 'module-name mod-name)))))
     ((and prelink? (not (member mod-name
				 (hfield-ref linker 'linked-interfaces))))
      (raise 'prelinked-body-interface-not-linked))
     ((and (eq? unit-type 'interface)
	   (member mod-name (hfield-ref linker 'linked-interfaces)))
      (begin
	(dwl1 "revisiting")
	'()))
     ((and (eq? unit-type 'body)
	   (member mod-name (hfield-ref linker 'linked-bodies)))
      (begin
	(dwl1 "revisiting")
	'()))
     ((and (memq unit-type '(proper-program script))
	   (not program?))
      (raise 'invalid-program-linkage))
     (else
      (dwl4 "theme-do-link/1")
      (dwl4 "new module")
      (let* ((ext (get-pcode-suffix unit-type))
	     (source-filename1
	      (if (string-null? source-filename)
		  (search-file (hfield-ref linker 'module-search-path)
			       mod-name ext)
		  source-filename)))
	(dwl4 "theme-do-link/2")
	(if (not (eqv? source-filename1 #f))
	    (let ((source-file
		   (theme-open-input-file source-filename1)))
	      (dwl4 "theme-do-link/3")
	      (let* ((name (theme-read source-file))
		     (actual-name (if (symbol? name) (list name) name))
		     (script0? (theme-read source-file))
		     (first-free-loc (theme-read source-file))
		     ;; Read imports, imports-with-reexports, and uses.
		     (imports
		      (append (theme-read source-file)
			      (theme-read source-file)
			      (theme-read source-file)))
		     (l-prelinked-bodies
		      (theme-read source-file)))
		(dwl4 "theme-do-link/4")
		(if (not (module-name=? mod-name actual-name))
		    (raise 'module-name-mismatch))
		;; It is essential to update linker-interfaces and linked-bodies
		;; here before further linking of the unit.
		(cond
		 ((eq? unit-type 'interface)
		  (hfield-set! linker 'linked-interfaces
			       (cons mod-name
				     (hfield-ref linker 'linked-interfaces))))
		 ((eq? unit-type 'body)
		  (hfield-set! linker 'linked-bodies
			       (cons mod-name
				     (hfield-ref linker 'linked-bodies)))))
		(let ((visited-cur (cons mod-name hierarchy)))
		  (dw1 "Linking unit imports: ") 
		  (dwl1 mod-name)
		  (cond
		   ((memq unit-type '(proper-program script))
		    (theme-link-program-imports linker address-env imports))
		   ((eq? unit-type 'interface)
		    (let ((new-hierarchy (cons mod-name hierarchy)))
		      (theme-link-interface-imports linker address-env new-hierarchy imports)))
		   ((eq? unit-type 'body)
		    (theme-link-body-imports linker address-env mod-name imports))
		   (else (raise 'internal-error)))
		  (if (not-null? l-prelinked-bodies)
		      (if (not (eq? unit-type 'interface))
			  (begin
			    (dw1 "Prelinking bodies: ")
			    (dwl1 mod-name)
			    (theme-prelink linker address-env visited-cur
					   l-prelinked-bodies))
			  (raise 'prelink-in-interface)))

		  (dw1 "Reading unit contents: ") 
		  (dwl1 mod-name)

		  ;; Updating variable numbering.
		  (let ((ht (hfield-ref linker 'ht-module-indices)))
		    ;; An interface is always read before its body
		    ;; so we do not need to handle bodies here.
		    (if (and (or (eq? unit-type 'interface)
				 (eq? unit-type 'proper-program)
				 (eq? unit-type 'script))
			     (not (hash-ref ht actual-name)))
			(let ((i-current-var-start
			       (hfield-ref linker 'i-next-var-start))
			      (i-number-of-vars
			       (if (eq? unit-type 'interface)
				   (get-first-free-loc linker actual-name)
				   first-free-loc)))
			    (hfield-set!
			     linker
			     'i-next-var-start
			     (+ (hfield-ref linker 'i-next-var-start)
				i-number-of-vars))
			    (hash-set! ht actual-name i-current-var-start))))

		  (let ((body
			 (theme-read-file source-file)))
		    (theme-close-input-port source-file)
		    (if (not-null? (cdr body)) (raise 'extra-code-in-impl-file))
		    (hfield-set! linker 'state 'target-compilation)
		    (dw1 "Target compilation: ")
		    (dw1 mod-name)
		    (dw1 ", ")
		    (dwl1 unit-type)
		    (do ((s-body (car body) (cdr s-body)))
			((null? s-body))
		      (let ((sexpr (car s-body)))
;;			(dw1 "compiling pcode expression: ")
;;			(dwl4 sexpr)
			(hfield-set! linker 'current-toplevel-expr sexpr)
			(let ((rexpr (theme-read-body-expr linker address-env sexpr)))
			  (hfield-set! linker 'current-toplevel-expr '())
			  (hfield-set! linker 'repr-list
				       (cons rexpr (hfield-ref linker 'repr-list))))))
			  ;; We should change the linking phase here.
			  ;; Field current-toplevel-expr should also retain
			  ;; its value during target compilation.
			  ;; (let ((texprs (theme-target-compile-body-repr linker rexpr)))
			  ;;   ;; texprs is a list consisting of target expressions
			  ;;   (guard (exc (else
			  ;; 		 (make-file-exception
			  ;; 		  'error-writing-file
			  ;; 		  (hfield-ref linker 'target-filename))))
			  ;; 	   (write-module-body
			  ;; 	    (hfield-ref linker 'target-file) texprs
			  ;; 	    (hfield-ref linker 'pretty-print?)))))))
		    (for-each
		     (lambda (imp)
		       (if (not (member imp (hfield-ref linker 'linked-bodies)))
			   (theme-link-unit linker "" imp 'body address-env
					    '() #f #f)))
		     imports)))))
	    (raise 'module-not-found)))))
    (hfield-set! linker 'current-module prev-module)))


(define gl-linker
  (make-hrecord <linker>
		'()
		#t
		#t
		#f
		#f
		#f
		#t
		#f
		#t
		#t
		'tree-il-2.2
		'()
		'()
		'()
		'()
		'()
		'()
		'()
		'()
		'()
		'()
		'()
		'()
		'()
		'()
		'()
		'()
		'()
		'()
		'()
		""
		""
		'()
		'()
		'()
		'()
		'()
		'()
		0
		0
		'()
		'()
		'()
		'()
		'()
		'()
		'()
		'()
		#f
		#f
		'()
		'()
		0
		'()
		'()
		'()
		'()
		'()
		'()))


(set! gl-linker-fwd gl-linker)


(define (init-linker linker compiler-root-env module-search-path
		     all-assertions? factorize? strip?
		     pretty-print?
		     verbose-errors?
		     verbose-typechecks?
		     verbose-unlinked-procedures?
		     backtrace?
		     runtime-pretty-backtrace?
		     s-intermediate-language)
  (assert (hrecord-is-instance? linker <linker>))
  (assert (hrecord-is-instance? compiler-root-env <environment>))
  (assert (and (list? module-search-path)
	       (and-map? string? module-search-path)))
  (assert (boolean? all-assertions?))
  (assert (boolean? pretty-print?))
  (assert (boolean? verbose-errors?))
  (assert (boolean? verbose-typechecks?))
  (assert (boolean? backtrace?))
  (assert (memq s-intermediate-language gl-l-all-int-lang))
  (let* ((linker-root-env (make-address-environment compiler-root-env))
	 (alloc-loc (lambda (s-name toplevel?)
		      (linker-alloc-loc linker s-name toplevel?)))
	 (param-cache-parsing (make-hrecord <param-cache> '()))
	 (param-cache-instantiation (make-hrecord <param-cache> '()))
	 (ht-used (make-hash-table gl-i-ht-used-size))
	 (ht-method-defs (make-hash-table gl-i-ht-method-defs-size))
	 (ht-method-decls (make-hash-table gl-i-ht-method-decls-size))
	 (ht-used-decls (make-hash-table gl-i-ht-used-decls-size))
	 (ht-rebound (make-hash-table gl-i-ht-rebound-size))
	 (ht-decl-types (make-hash-table gl-i-ht-decl-types-size))
	 (ht-globals-by-address
	  (make-hash-table gl-i-ht-globals-by-address-size))
	 (ht-globals-by-name
	  (make-hash-table gl-i-ht-globals-by-name-size))
	 (ht-cycles (make-hash-table gl-i-ht-cycles-size))
	 (ht-goops-classes (make-hash-table gl-i-ht-goops-classes-size))
	 (ht-prim-classes (make-hash-table gl-i-ht-prim-classes-size))
	 (ht-fact (make-hash-table gl-i-ht-fact-size))
	 (ht-module-indices (make-hash-table gl-i-ht-module-indices-size))
	 (ht-lexical-vars (make-hash-table gl-i-ht-lexical-vars-size))
	 (ht-raw-procs (make-hash-table gl-i-ht-raw-procs-size))
	 (ht-equal (make-hash-table gl-i-ht-equal-size))
	 (ht-equal-objects (make-hash-table gl-i-ht-equal-objects-size))
	 (ht-equal-contents (make-hash-table gl-i-ht-equal-contents-size))
	 (optimize-raw-proc-inst?
	  (if (memq s-intermediate-language gl-l-tree-il-lang)
	      #t
	      #f))
	 (binder-parsing (make-hrecord <binder>
				       param-cache-parsing
				       alloc-loc
				       #t
				       #t
				       #f
				       #f
				       #f
				       #f
				       #f
				       '()
				       ht-globals-by-address
				       ht-method-decls
				       '()
				       '()
				       '()
				       #f
				       #f
				       '()
				       '()
				       ""
				       '()
				       optimize-raw-proc-inst?
				       ht-raw-procs))
	 (binder-instantiation (make-hrecord <binder>
					     param-cache-instantiation
					     alloc-loc
					     #t
					     #t
					     #f
					     #t
					     #t
					     #f
					     #f
					     '()
					     ht-globals-by-address
					     ht-method-decls
					     '()
					     '()
					     '()
					     #f
					     #f
					     '()
					     '()
					     ""
					     '()
					     optimize-raw-proc-inst?
					     ht-raw-procs)))
    (address-env-add-binding! linker-root-env tp-cast-vector0)
    (address-env-add-binding! linker-root-env tp-cast-mutable-vector0)
    (address-env-add-binding! linker-root-env tp-cast-value-vector0)
    (address-env-add-binding! linker-root-env tp-cast-mutable-value-vector0)
    (hfield-set! linker 'state 'initial)
    (hfield-set! linker 'all-assertions? all-assertions?)
    (hfield-set! linker 'factorize? factorize?)
    (hfield-set! linker 'pretty-print? pretty-print?)
    (hfield-set! linker 'verbose-errors? verbose-errors?)
    (hfield-set! linker 'verbose-typechecks? verbose-typechecks?)
    (hfield-set! linker 'backtrace? backtrace?)
    (hfield-set! linker 'runtime-pretty-backtrace?
		 runtime-pretty-backtrace?)
    (hfield-set! linker 'strip? strip?)
    (hfield-set! linker 'verbose-unlinked-procedures?
		 verbose-unlinked-procedures?)
    (hfield-set! linker 's-intermediate-language
		 s-intermediate-language)
    (hfield-set! linker 'repr-list '())
    (hfield-set! linker 'ht-used ht-used)
    (hfield-set! linker 'ht-method-defs ht-method-defs)
    (hfield-set! linker 'ht-method-decls ht-method-decls)
    (hfield-set! linker 'ht-used-decls ht-used-decls)
    (hfield-set! linker 'ht-rebound ht-rebound)
    (hfield-set! linker 'ht-decl-types ht-decl-types)
    (hfield-set! linker 'ht-cycles ht-cycles)
    (hfield-set! linker 'ht-goops-classes ht-goops-classes)
    (hfield-set! linker 'ht-prim-classes ht-prim-classes)
    (hfield-set! linker 'ht-fact ht-fact)
    (hfield-set! linker 'ht-module-indices ht-module-indices)
    (hfield-set! linker 'ht-lexical-vars ht-lexical-vars)
    (hfield-set! linker 'ht-equal ht-equal)
    (hfield-set! linker 'ht-equal-objects ht-equal-objects)
    (hfield-set! linker 'ht-equal-contents ht-equal-contents)
    (hfield-set! linker 'lst-enclosing-cycles '())
    (hfield-set! linker 'interm-file '())
    (hfield-set! linker 'interm-filename "")
    (hfield-set! linker 'target-filename "")
    (hfield-set! linker 'module-search-path module-search-path)
    (hfield-set! linker 'param-cache-parsing param-cache-parsing)
    (hfield-set! linker 'param-cache-instantiation param-cache-instantiation)
    (hfield-set! linker 'binder-parsing binder-parsing)
    (hfield-set! linker 'binder-instantiation binder-instantiation)
    (hfield-set! linker 'fixed-tvars '())
    (hfield-set! linker 'next-free-loc gl-i-loc-start)
    (hfield-set! linker 'next-tvar-number 0)
    (hfield-set! linker 'current-module '())
    (hfield-set! linker 'linked-interfaces '())
    (hfield-set! linker 'linked-bodies '())
    (hfield-set! linker 'root-env linker-root-env)
    (hfield-set! linker 'global-decls '())
    (hfield-set! linker 'ht-globals-by-name ht-globals-by-name)
    (hfield-set! linker 'ht-globals-by-address ht-globals-by-address)
    (hfield-set! linker 'decl-proc-instances '())
    (hfield-set! linker 'inside-param-def? #f)
    (hfield-set! linker 'tcomp-inside-param-proc? #f)
    (hfield-set! linker 't-tvars '())
    (hfield-set! linker 'visited-in-binding '())
    (hfield-set! linker 'i-next-var-start 0)
    (hfield-set! linker 'current-expr '())
    (hfield-set! linker 'current-toplevel-expr '())
    (hfield-set! linker 'current-repr '())
    (hfield-set! linker 'current-toplevel-repr '())
    (hfield-set! linker 'current-instance '())
    (hfield-set! linker 'current-repr-to-bind '())))


(define (make-extra-initializations linker)
  (assert (hrecord-is-instance? linker <linker>))
  (let ((first-free-number (hfield-ref linker 'next-tvar-number)))
    (if (use-tree-il? linker)
	(let ((s-call (tree-il-get-call-keyword linker)))
	  `((,s-call (toplevel vector-set!)
		     (,s-call (toplevel vector-ref)
			      (toplevel gl-rte)
			      (toplevel i-rte-tvar-allocator))
		     (toplevel i-tva-start)
		     (const ,first-free-number))))
	`((vector-set! (vector-ref gl-rte i-rte-tvar-allocator)
		       i-tva-start ,first-free-number)))))


(define (is-my-main? address)
  (and
   (eqv? (hfield-ref address 'number) address-number-target)
   (eqv? (hfield-ref address 'source-name) '_main)))


(define (get-main linker)
  (address-env-get-item-general
   (hfield-ref linker 'root-env)
   (lambda (key value) (is-my-main? key))))


(define (get-main-def lst-reprs)
  (find (lambda (repr)
	  (and
	   (hrecord-is-instance? repr <variable-definition>)
	   (is-my-main? (hfield-ref (hfield-ref repr 'variable) 'address))))
	lst-reprs))


(define (compile-main linker)
  (if (use-tree-il? linker)
      (tc-tree-il-define-main linker)
      (tc-scheme-define-main linker)))


(define (theme-link-program linker program-name
			    source-filename
			    interm-filename
			    target-filename
			    module-search-path
			    s-intermediate-language
			    final-compilation?
			    keep-intermediate?
			    link-to-cache?
			    all-assertions?
			    factorize?
			    strip?
			    pretty-print?
			    verbose-errors?
			    verbose-typechecks?
			    verbose-unlinked-procedures?
			    backtrace?
			    runtime-pretty-backtrace?)
  (assert (hrecord-is-instance? linker <linker>))
  (assert (or (symbol? program-name) (is-module-name? program-name)))
  (assert (and (string? source-filename) (not (string-null? source-filename))))
  (assert (string? interm-filename))
  (assert (string? target-filename))
  (assert (boolean? final-compilation?))
  (assert (boolean? all-assertions?))
  (assert (boolean? factorize?))
  (assert (boolean? strip?))
  (assert (boolean? pretty-print?))
  (assert (boolean? verbose-errors?))
  (assert (boolean? verbose-typechecks?))
  (assert (boolean? verbose-unlinked-procedures?))
  (assert (boolean? backtrace?))
  (assert (boolean? runtime-pretty-backtrace?))
  (assert (memq s-intermediate-language gl-l-all-int-lang))
  (dw1 "Linking program ")
  (dwl1 program-name)
  (guard
   (exc (else (handle-linker-error linker exc)))
   (let* ((l-comp '())
	  (interm-file-ext (get-interm-file-ext s-intermediate-language))
	  (interm-filename1
	   (if (string-null? interm-filename)
	       (string-append
		(get-filename-without-ext
		 (get-filename-without-path source-filename))
		interm-file-ext)
	       interm-filename))
	  (interm-file (theme-open-output-file interm-filename1))
	  (script? (eq? (get-pcode-unit-type source-filename) 'script))
	  (target-filename1
	   (if (and final-compilation? (not link-to-cache?))
	       (if (string-null? target-filename)
		   (string-append
		    (get-filename-without-ext
		     (get-filename-without-path source-filename))
		    target-file-ext)
		   target-filename)
	       "")))
     (init-linker linker global-builtins-symtbl
		  module-search-path all-assertions? factorize? strip?
		  pretty-print? verbose-errors?
		  verbose-typechecks?
		  verbose-unlinked-procedures?
		  backtrace?
		  runtime-pretty-backtrace?
		  s-intermediate-language)
     (set! gl-binder (hfield-ref linker 'binder-parsing))
     (hfield-set! linker 'interm-file interm-file)
     (hfield-set! linker 'interm-filename interm-filename1)
     (hfield-set! linker 'target-filename target-filename1)
     (let ((l-prefix
	    (if (use-tree-il? linker)
		(compile-program-prefix-tree-il linker script?)
		(compile-program-prefix-scheme script?))))
       (set! l-comp l-prefix))
     (hfield-set! linker 'state 'pcode-reading)
     (theme-link-unit linker source-filename program-name
		      (if script? 'script 'proper-program)
		      (hfield-ref linker 'root-env) '() #t #f)
     (if (and (not script?)
	      (not (get-main linker)))
	 (raise 'main-procedure-not-defined))
     (set! l-comp
	   (append l-comp (compile-decl-param-proc-instances linker)))
     (set! l-comp
	   (append
	    l-comp
	    (make-extra-initializations linker)))
     (let ((repr-list (hfield-ref linker 'repr-list)))
       (if (hfield-ref linker 'strip?)
	   (begin
	     (dwl1 "Performing coverage analysis")
	     (hfield-set! linker 'state 'coverage-analysis)
	     (let ((ht-var-defs (make-var-def-hash-table repr-list)))
	       (hfield-set! linker 'ht-var-defs ht-var-defs)
	       (determine-total-coverage linker)
	       (do-prevent-stripping linker))))
       (hfield-set! linker 'state 'instantiation)
       (set! gl-binder (hfield-ref linker 'binder-instantiation))
       (dwl1 "Computing parametrized entity instances")
       (dw1 "Number of expressions to handle: ")
       (dwl1 (length (hfield-ref linker 'repr-list)))
       (let ((lst-processed
	      (theme-instantiate-program linker
					 (hfield-ref linker 'repr-list))))
	 (hfield-set! linker 'state 'target-compilation)
	 (dwl1 "Linking the target entities")
	 (set! l-comp
	       (append l-comp
		       (theme-target-compile-program linker lst-processed)))
	 (set! l-comp
	       (append l-comp
		       (compile-used-method-decls linker)))))
     (if (not script?)
	 (set! l-comp (append l-comp (list (compile-main linker)))))
     (if (use-tree-il? linker)
	 (let ((l-comp2
		(tc-tree-il-compile-sequence
		 linker
		 (filter-void (handle-tree-il-splices linker l-comp)))))
	   (if pretty-print?
	       (pretty-print l-comp2 interm-file)
;;	       (pretty-print l-comp interm-file)
	       (write l-comp2 interm-file))
	   (newline interm-file))
	 (for-each (lambda (texpr)
		     (write-target-expr linker interm-file pretty-print?
					texpr))
		   l-comp))
     (theme-close-output-port interm-file)
     (hfield-set! linker 'interm-file '())
     ;; We keep the intermediate filename field during the final compilation.
     (if final-compilation?
	 (begin
	   (dwl1 "Compiling the guile objcode file")
	   (hfield-set! linker 'state 'final-compilation)
	   (if link-to-cache?
	       (if (not (= (system
			    (if (use-tree-il? linker)
				(string-append "guild compile --from=tree-il "
					       interm-filename1)
				(string-append "guild compile "
					       interm-filename1)))
			   0))
		   (raise 'error-in-final-compilation))
	       ;; (dwl1 "xxx")
	       ;; (dwl1 target-filename)
	       ;; (dwl1 target-filename1)
	       (if (not (= (system
			    (if (use-tree-il? linker)
				(string-append "guild compile "
					       "--from=tree-il --output="
					       target-filename1
					       " "
					       interm-filename1)
				(string-append "guild compile --output="
					       target-filename1
					       " "
					       interm-filename1)))
			   0))
		   (raise 'error-in-final-compilation)))))
     (hfield-set! linker 'interm-filename "")
     (hfield-set! linker 'target-filename "")
     (if (not keep-intermediate?)
	 (delete-file interm-filename1))))
  (write-line-info "Program linked successfully."))


;; Note that theme-link-program1 puts the target file
;; into the same directory with the source file.
;; This file is to be used inside guile interpreter
;; so we set backtrace on.
(define (theme-link-program1 main-module module-search-path
			     s-intermediate-language
			     script?
			     final-compilation?
			     keep-intermediate?
			     link-to-cache?
			     factorize?
			     strip?
			     all-assertions?
			     pretty-print?
			     verbose-errors?
			     verbose-typechecks?
			     verbose-unlinked-procedures?
			     runtime-pretty-backtrace?)
  (assert (or (symbol? main-module) (is-module-name? main-module)))
  (dwl3 "theme-link-program1")
  (dwl3 strip?)
  (let* ((actual-main-module (get-actual-module-name main-module))
	 (ext (if script?
		  (get-pcode-suffix 'script)
		  (get-pcode-suffix 'proper-program)))
	 (source-filename (search-file module-search-path
				       actual-main-module ext)))
    (if (not (eq? source-filename #f))
	(let* ((basename (get-filename-without-ext source-filename))
	       (interm-file-ext (get-interm-file-ext s-intermediate-language))
	       (interm-filename (string-append basename interm-file-ext))
	       (target-filename (string-append basename target-file-ext)))
	  (theme-link-program gl-linker actual-main-module
			      source-filename interm-filename target-filename
			      module-search-path
			      s-intermediate-language
			      script?
			      final-compilation?
			      keep-intermediate?
			      link-to-cache?
			      all-assertions?
			      factorize?
			      strip?
			      pretty-print?
			      verbose-errors?
			      verbose-typechecks?
			      verbose-unlinked-procedures?
			      #t
			      runtime-pretty-backtrace?))
	(raise 'program-file-not-found))))
